#!/usr/bin/env python
"""

Base 100 floating point prototype
Copyright 2014 Damian Yerrick

Copying and distribution of this file, with or without
modification, are permitted in any medium without royalty
provided the copyright notice and this notice are preserved.
This file is offered as-is, without any warranty.

"the determined Real Programmer can write Fortran programs in any language."
--Ed Post, Real Programmers Don't Use Pascal, 1982
http://www.pbm.com/~lindahl/real.programmers.html

This is a prototype of a base-100 floating point library that I
plan to implement in 6502 assembly language.  It treats a number as
a 12-digit coefficient times 100 raised to a positive or negative
power.  The coefficient is stored as a sequence of six pairs of
digits from 0 to 99.

Known limits as of today:
* No sign bit (as my target application doesn't need negative and
  positive numbers in the same context)
* No floating-point division or square root (likewise not needed for
  my app)
* Ill-defined behavior on overflow during addition or multiplication
  when numbers exceed a quadragintillion (10^123)

Future:
* Have some other smart people play with it to see what I messed up
* Rewrite in 6502 assembly
* Triangular roots
* Implement all the rules of Cookie Clicker

Changelog:
2014-03-27
* initial release

"""
from __future__ import print_function, with_statement, division

# Table of quarter squares, used in multiplication
qsqlo = [(x - 100)*(x - 100)//4 for x in xrange(300)]
qsqhi = [x // 100 for x in qsqlo]
qsqlo = [x % 100 for x in qsqlo]

si_pfxs = "afpnum kMGTPEZY"
decpoint = '.'
thousandspoint = ','

# A floating point number is 7 bytes, the first 6 of which contain
# digit pairs (0-99) and the last of which is an exponent (times
# 100 to the nth power).

def dfpmake(i, exponent=0):
    a = [0, 0, 0, 0, 0, 0, exponent + 127]
    while i > 0:
        a[6] += 1
        if a[5] >= 50:
            a[4] += 1  # potential bias: always rounds half up
        a[1:6] = a[0:5]
        a[0] = i % 100
        i //= 100
    return tuple(a)

def dfpfmt(a):
    """Format a number as a string."""
    # Print the whole number part
    wholedp = a[6] - 127  # number of digit pairs in whole part
    digs = []
    if wholedp > 0:
        for i in xrange(min(wholedp, 6)):
            dig = a[i]
            for dig in dig // 10, dig % 10:
                if digs or dig:
                    digs.append(chr(dig | 0x30))
        if wholedp > 6:
            digs.append('00' * (wholedp - 6))
    if not digs:
        digs.append('0')

    # Look for a fractional part
    last = 5
    while last >= max(wholedp, 0):
        if a[last]:
            break
        last -= 1
    last += 1

    # Print the fractional part        
    if last > 0 and last > wholedp:
        digs.append(decpoint)
        digs.append('00' * -wholedp)
        for i in xrange(max(0, wholedp), last):
            dig = a[i]
            for dig in dig // 10, dig % 10:
                digs.append(chr(dig | 0x30))
        if digs[-1] == '0':
            del digs[-1]

    return ''.join(digs)

def dfpnzrange(a):
    """Find the smallest slice of nonzero coefficient digit pairs."""
    firstnz = 0
    while firstnz < 6 and not a[firstnz]:
        firstnz += 1
        if firstnz >= 6:
            return 6, 0
    lastnz = 5
    while lastnz >= firstnz and not a[lastnz]:
        lastnz -= 1
    return firstnz, lastnz + 1

def dfpfmt_metric(a, maxlen):
    """Format a DFP as a string, using metric prefixes to fit in a given length.

maxlen must be at least 6

"""
    i, lastnz = dfpnzrange(a)
    if i >= 6:
        return '0'
    wholedigits = (a[6] - i - 127) * 2
    if a[i] < 10:
        wholedigits -= 1
    tillcomma = wholedigits % 3 or 3

    # Forms that don't need prefixes:
    # maxlen  wholedigits  example
    #   6       5              99,999
    #   7       6             999,999
    #   8       6             999,999
    #   9       7           9,999,999
    #  10       8          99,999,999
    # Decimal point doesn't need prefixes if fewer than 3 leading
    # zeroes.  Examples with maxlen=7:
    # wholedigits  example
    #    0         .999999
    #   -1         .099999
    #   -2         .009999
    #   -3         999.9 u
    #   -4         99.99 u
    # If wholedigits is at least -2 and less than the maximum that
    # will fit in maxlen, format them normally.
    if not -2 <= wholedigits <= ((maxlen + 1) * 3 // 4):
        pfxtype = si_pfxs[(wholedigits + 17) // 3]
        maxlen -= 2
        wholedigits = tillcomma
    else:
        pfxtype = None

    digs = []
    lastwholedigit = -1
    while wholedigits > 0:
        dp = a[i] if i < 6 else 0
        dp = (dp // 10, dp % 10)
        # If the decimal point lies between digits of a pair, which
        # may happen with nano, milli, kilo, or giga prefixes,
        # save the second digit
        for dig in dp:
            if wholedigits <= 0:
                lastwholedigit = dig
            elif dig or digs:
                if tillcomma <= 0:
                    tillcomma = 3
                    digs.append(thousandspoint)
                digs.append(chr(48 + dig))
                tillcomma -= 1
                wholedigits -= 1
        i += 1
    # If digits past decimal point exist, print them
    if i < lastnz and len(digs) + 2 <= maxlen:
        digs.append(decpoint)
        if lastwholedigit >= 0:
            digs.append(chr(48 + lastwholedigit))
        while i < lastnz and len(digs) < maxlen:
            dp = a[i]
            i += 1
            hdig, ldig = dp // 10, dp % 10
            digs.append(chr(48 + hdig))
            if len(digs) < maxlen and (i < lastnz or ldig > 0):
                digs.append(chr(48 + ldig))

    if pfxtype:
        digs.extend((' ', pfxtype))
    return ''.join(digs)

def dfpnorm(a):
    """Return normalized a."""
    firstnz = dfpnzrange(a)[0]
    if firstnz >= 6:
        return (0, 0, 0, 0, 0, 0, 0)
    # Significant digits, trailing zeroes to match removed leading
    # zeroes, and the new exponent
    b = list(a[firstnz:6]) + [0] * firstnz + [a[6] - firstnz]
    return tuple(b)

def dfpinc(a, startidx=5):
    """Increment a DFP as a list by one unit starting at digitpair idx."""
    while startidx >= 0:
        a[startidx] += 1
        if a[startidx] < 100:
            return 0
        a[startidx] -= 100
        startidx -= 1
    return 1

def dfpdouble(a, carry=0):
    for i in xrange(5, -1, -1):
        dp = a[i] * 2 + carry
        carry = 1 if dp >= 100 else 0
        a[i] = dp - (100 if carry else 0)
    return carry

def dfphalve(a, carry=0):
    for i in xrange(6):
        dp = a[i] + (100 if carry else 0)
        carry = dp & 1
        a[i] = dp // 2
    return carry

def dfpadd(a, b):
    """Return the sum of a and b."""
    # Make `a` have a larger exponent without loss of generality.
    if a[6] < b[6]:
        a, b = b, a
    a, b = list(a), list(b)

    # At this point, `a` has a larger exponent.
    # Start by shifting `a` as far left as needed and possible.
    if a[6] > b[6] and not a[0]:
        shiftamt, aend = dfpnzrange(a)
        shiftamt = min(shiftamt, a[6] - b[6])
        a[6] -= shiftamt
        a[:6 - shiftamt] = a[shiftamt:6]
        a[6 - shiftamt:6] = [0] * shiftamt

    # If the exponents differ after fronting a, we must shift b,
    # inexactitude be damned.  Round for the first time.
    if a[6] > b[6]:
        shiftamt = a[6] - b[6]
        # Set initial carry based on how b is rounded
        carry = 1 if shiftamt <= 6 and b[6 - shiftamt] > 50 else 0
        if shiftamt >= 6 and not carry:
            return tuple(a)
        b[shiftamt:6] = b[:6 - shiftamt]
        b[:shiftamt] = [0] * shiftamt
        b[6] = a[6]
    else:
        carry = 0

    # Add the numbers digit pair by digit pair.
    for i in xrange(5, -1, -1):
        dp = a[i] + b[i] + carry
        carry = 1 if dp >= 100 else 0
        a[i] = dp - (100 if carry else 0)

    # If carry out (e.g. 50 + 50), round again.
    if carry:
        if a[5] >= 50:
            carry = dfpinc(a, 4)
            assert carry == 0
        a[1:6] = a[0:5]
        a[0] = 1
        a[6] += 1
    return tuple(a)

def dfpsub(a, b, signonly=False):
    # First compare the values to find the result's sign
    a = dfpnorm(a)
    b = dfpnorm(b)
    if b[6] > a[6]:
        sign = -1
    elif b[6] < a[6]:
        sign = 1
    elif b[6] == a[6]:

        # same exponent; compare digit-pairwise
        sign = 0
        for adp, bdp in zip(a[:6], b[:6]):
            if adp > bdp:
                sign = 1
                break
            elif adp < bdp:
                sign = -1
                break
    if signonly:
        return sign
    if sign == 0:
        return (0, 0, 0, 0, 0, 0, 0), 0
    if sign < 0:
        a, b = b, a

    # Check for essentially subtracting zero
    expdiff = a[6] - b[6]
    assert expdiff >= 0
    if expdiff > 6 or (expdiff == 6 and b < 50):
        return a, sign

    # If like exponents, use ten's complement addition
    a = list(a)
    if expdiff == 0:
        print("Like exponents")
        carry = 1
        for i in range(5, -1, -1):
            dp = 99 + carry + a[i] - b[i]
            carry = 1 if dp >= 100 else 0
            a[i] = dp - (100 if carry else 0)
        assert carry
        return tuple(a), sign

    # One guard digit pair may be needed.  Shift b to use an exponent
    # one less than a, and set initial carry based on rounding
    b = list(b)
    shiftamt = expdiff - 1
    # Set initial carry based on how b is rounded
    carry = 2 if b[6 - shiftamt] >= 50 else 1
    b[shiftamt:6] = b[:6 - shiftamt]
    b[:shiftamt] = [0] * shiftamt
    b[6] += shiftamt
    a[6] = 0
##    print("Unlike exponents: b is now %s and carry is %d\n%s"
##          % (dfpfmt(b), carry, repr(b)))

    for i in range(5, -1, -1):
        dp = 99 + carry + a[i + 1] - b[i]
        carry = 1 if dp >= 100 else 0
        a[i + 1] = dp - (100 if carry else 0)
    a[0] += carry - 1
    assert a[0] >= 0

    # If subtraction didn't lose the high digit pair, round.
    if a > 0:  # DUBIOUS
        if a[6] >= 50:
            dfpinc(a, 5)
        a[6] = b[6] + 1
    else:
        a[0:6] = a[1:7]
        a[6] = b[6]
    return tuple(a), sign

def dfpcmp(a, b):
    """Return the sign of b - a."""
    return dfpsub(a, b, true)

def dfpmul(a, b):
    c = [0]*7
    # normalizing first preserves precision
    a, b = dfpnorm(a), dfpnorm(b)
    cexp = a[6] + b[6] - 127
    if cexp <= 0:
        return (0, 0, 0, 0, 0, 0, 0)
    for adp in a[5::-1]:
        # This represents eight bytes of pointers on zero page
        c[1:7] = c[0:6]
        c[0] = 0
        if not adp:
            continue
        qsqplus = 100 + adp
        qsqminus = 100 - adp
        for i in xrange(5, -1, -1):
            bdp = b[i]
            shi = qsqhi[qsqplus + bdp] - qsqhi[qsqminus + bdp]
            assert shi >= 0
            slo = qsqlo[qsqplus + bdp] - qsqlo[qsqminus + bdp]
            if slo < 0:
                assert shi >= 1
                slo += 100
                shi -= 1
            if slo >= 100:
                slo -= 100
                shi += 1
            assert adp * bdp == shi * 100 + slo

            slo += c[i + 1]
            if slo >= 100:
                slo -= 100
                shi += 1
            c[i + 1] = slo
            shi += c[i]
            if shi >= 100:
                shi -= 100
                dfpincresult = dfpinc(c, i - 1)
                assert dfpincresult == 0
            c[i] = shi
    if c[0]:
        if c[6] >= 50:
            dfpinc(c, 5)
    else:
        c[0:6] = c[1:7]
        cexp -= 1
    if cexp <= 0:
        return (0, 0, 0, 0, 0, 0, 0)
    c[6] = cexp
    return tuple(c)

def dfppayoff(cost, cps):
    """Calculates the payoff time for cost A and rate B."""

    cost = dfpnorm(cost)
    cps = dfpnorm(cps)
    if cost[6] < cps[6]:
        return (0, 0, 0, 0)
##    print("cost is %s" % repr(cost))
##    print("cps is %s" % repr(cps))

    # Find a power of 2 that is greater than cost/cps.  Start by
    # multiplying cps by large powers until exponent matches that
    # of cost.
    bits = 0
    f64 = dfpmake(64)
    while cost[6] >= cps[6] + 2:
        bits += 13
        cps = dfpmul(cps, dfpmake(8192))
##        print("cps: %s/2^%d" % (repr(cps), bits))
    while cost[6] > cps[6]:
        bits += 6
        cps = dfpmul(cps, dfpmake(64))
##        print("cps: %s/2^%d" % (repr(cps), bits))

    # Make sure cost mantissa < cps mantissa
    cost = list(cost)
    cps = list(cps)
    while cps[0] <= 50 and cps[0] <= cost[0]:
        bits += 1
        dfpdouble(cps)
##        print("cps: %s/2^%d" % (repr(cps), bits))
    if cost[0] >= cps[0]:
        assert cost[0]
        bits += 1
        dfphalve(cost)
##        print("cost: %s*2" % repr(cost))

    # Now do binary long division: double cost until it exceeds cps
    days, hours, minutes, seconds = 0, 0, 0, 0
    while bits > 0:
        # Double cost
        carry = dfpdouble(cost)
##        print("cost is now %d,%s; " % (carry, repr(cost)), end='')

        # Trial subtraction: use element by element comparison
        carry = carry or cost[:6] >= cps[:6]
##        print("carry" if carry else "no carry")

        if carry:
            # Subtract coeffs only
            for i in range(5, -1, -1):
                dp = 99 + carry + cost[i] - cps[i]
                carry = 1 if dp >= 100 else 0
                cost[i] = dp - (100 if carry else 0)
            carry = 1
        seconds = seconds * 2 + carry
        carry = 1 if seconds >= 60 else 0
        if carry: seconds -= 60
        minutes = minutes * 2 + carry
        carry = 1 if minutes >= 60 else 0
        if carry: minutes -= 60
        hours = hours * 2 + carry
        carry = 1 if hours >= 24 else 0
        if carry: hours -= 24
        days = days * 2 + carry
        bits -= 1
    return (days, hours, minutes, seconds)

def test_dfpfmt():
    print(dfpmake(11))
    i = dfpmake(12345)
    i = dfpmake(1234500, -1)
    print(dfpfmt(i))
    print("nzrange of %s is %s" % (i, dfpnzrange(i)))
    print(dfpfmt(dfpmake(11)))
    print(dfpfmt(dfpmake(4440000)))
    print(dfpfmt(dfpmake(12345678987654321)))
    print(dfpfmt(dfpmake(11, -1)))
    print(dfpfmt(dfpmake(11, -2)))
    a = dfpmake(12)
    b = dfpmake(345)
    dfpadd(a, b)
    for i in (123, 1234, 12345, 123456, 1234567, 12345678, 123456789, 1234567890, 12345678901):
        i = dfpmake(i)
        print("\n".join("%s: %s" % (length, dfpfmt_metric(i, length))
                        for length in xrange(6, 15)))

def test_dfpadd():
    for i in (123, 1234, 12345, 123456, 1234567, 12345678, 123456789, 1234567890, 12345678901):
        print(i + 98.765)
        a_sum = dfpadd(dfpmake(i), dfpmake(987650, -2))
        print(dfpfmt(a_sum), dfpfmt_metric(a_sum, 15),
              dfpfmt_metric(a_sum, 7))

def test_dfpsub():
    for i in (123, 1234, 12345, 123456, 1234567, 12345678, 123456789, 1234567890, 12345678901):
        print(i - 456.78)
        a_diff, a_sign = dfpsub(dfpmake(i), dfpmake(45678, -1))
        tsign = "0+-"[a_sign]
        print("%s%s %s%s %s%s"
              % (tsign, dfpfmt(a_diff), tsign, dfpfmt_metric(a_diff, 15),
                 tsign, dfpfmt_metric(a_diff, 7)))

def calc_building_cost(base_cost, num_owned=0, santas=0, mul=dfpmul):
    """Calculates the cost of a building in Orteil's Cookie Clicker.

santas -- 0 for no Santa upgrades,
1 for Season savings or Santa's dominion, 2 for both

"""
    if santas:
        onepercentoff = dfpmake(99, -1)
        for i in xrange(santas):
            base_cost = mul(base_cost, onepercentoff)

    cost_ratio = dfpmake(115, -1)
    while True:
        if num_owned & 1:
            base_cost = mul(base_cost, cost_ratio)
        num_owned = num_owned // 2
        if num_owned == 0:
            return base_cost
        cost_ratio = mul(cost_ratio, cost_ratio)


def test_dfpmul():
    pairs = [
        (2, 3), (22, 33), (101, 1), (12345, 54321),
        (174900625, 1), (1, 174900625), (174900625, 174900625),
        ((2, 4), (3, -4)), ((2, -60), (3, -60)), ((2, -70), (3, -70)),
    ]
    for a, b in pairs:
        if not isinstance(a, tuple):
            a = (a, 0)
        two = dfpmake(*a)
        a = a[0] * (100 ** a[1])
        if not isinstance(b, tuple):
            b = (b, 0)
        three = dfpmake(*b)
        b = b[0] * (100.0 ** b[1])
        six = dfpmul(two, three)
        print("%g * %g = %g\n  so %s * %s = %s"
              % (a, b, a * b, dfpfmt(two), dfpfmt(three), dfpfmt(six)))
    print("Cost of cursors in Cookie Clicker")
    base_cost = dfpmake(15)
    for num in xrange(240, 247):
        cost = calc_building_cost(base_cost, num)
        print("Cursor while owning %d costs %s cookies\n(expect %d)"
              % (num, dfpfmt(cost), 15 * 1.15 ** num))
    print("246th cursor in game costs 12,561,178,154,785,838 with\n"
          "both Santa upgrades that affect building prices")
    print("in our library it costs %s"
          % dfpfmt(calc_building_cost(base_cost, num, 2)))

def test_dfppayoff():
    tcases = [
        (15, 10, -1), (100, 50, -1), (500, 4), # cursor, grandma, farm
        (3000, 10), (10000, 40), (40000, 100),  # factory, mine, shipment
        (200000, 400), (1666666, 6666),  # alchemy lab, portal
        (123456789, 98765), (3999999999, 999999),  # tm, ac
        (50000000000, 10, 3),  # prism
        (11646241140639, 10, 3)  # 10th prism
    ]
    for row in tcases:
        try:
            exponent = row[2]
        except IndexError:
            exponent = 0
        dfcost = dfpmake(row[0])
        dfcps = dfpmake(row[1], exponent)
        result = dfppayoff(dfcost, dfcps)
        tresult = "%dd %d:%02d:%02d" % result
        print("%d / %.2f = %s"
              % (row[0], row[1] * (100.0 ** exponent), tresult))

def main():
    test_dfpadd()
    test_dfpsub()
    test_dfpmul()
    test_dfppayoff()


if __name__=='__main__':
    main()
