#!/usr/bin/env python
from __future__ import with_statement, division
import os, sys, math, sim, floatypoint, array, subprocess

DFPREG_A, DFPREG_B, DFPREG_C, DFPREG_D = xrange(0, 28, 7)

# current contents of dfp_addresses
#  0: floattest_fpa
#  1: dfpfromu8
#  2: dfpload
#  3: dfpstore
#  4: dfpnzrange
#  5: dfpnorm
#  6: dfpformat_a
#  7: dfpbeautify_a
#  8: dfpdouble
#  9: dfphalve
# 10: dfpinc
# 11: dfpcmp
# 12: dfpadd
# 13: dfpsub
# 14: dfpmul
# 15: dfpdiv

def jsr(address, a=0, x=0, y=0, p=0, trace=False):
    regs = {'a': a, 'x': x, 'y': y, 'p': p, 's': 0xFD, 'pc': address}
    cycles = 0
    tracelog = []
    while regs['s'] <= 0xFD:
        lastpc = regs['pc']
        cycles += sim.run_instruction(regs)
        if trace:
            tracelog.append("%04x:%s a=%02x x=%02x y=%02x p=%02x s=%02x"
                            % (lastpc, sim.opcodes[regs['lastopcode']],
                               regs['a'], regs['x'], regs['y'], regs['p'], regs['s']))
    if tracelog:
        print("\n".join(tracelog))
    return regs, cycles

def strcpyfrom6502(address):
    "Copy a NUL-terminated string from 6502 RAM"
    strbuilder = array.array('B')
    while len(strbuilder) < 256:
        c = sim.ram[address + len(strbuilder)]
        if c == 0:
            break
        strbuilder.append(c)
    return strbuilder.tostring()

def dfpload6502(reg, num):
    sim.ram[256:256 + len(num)] = array.array('B', num)
    jsr(dfp_addresses[2], a=1, y=0, x=reg)

def dfpstore6502(reg):
    jsr(dfp_addresses[3], a=1, y=0, x=reg)
    return tuple(sim.ram[0x0100:0x0107])

def dfpnzrange6502(num):
    dfpload6502(DFPREG_A, num)
    regs, _ = jsr(dfp_addresses[4], x=DFPREG_A)
    return min(6, regs['a']), regs['y']

def dfpnorm6502(num):
    dfpload6502(DFPREG_A, num)
    regs, _ = jsr(dfp_addresses[5], x=DFPREG_A)
    return dfpstore6502(DFPREG_A)
    
def dfpfmt6502(num):
    dfpload6502(DFPREG_A, num)
    jsr(dfp_addresses[6], a=1, y=0)
    return strcpyfrom6502(0x0100)

def dfpfmt_metric6502(num, width):
    dfpload6502(DFPREG_A, num)
    jsr(dfp_addresses[7], a=1, y=0, x=width)
    return strcpyfrom6502(0x0100)

def dfpdouble6502(num, carry=0):
    dfpload6502(DFPREG_A, num)
    regs, _ = jsr(dfp_addresses[8], x=DFPREG_A, p=1 if carry else 0)
    return dfpstore6502(DFPREG_A), regs['p'] & 0x01

def dfphalve6502(num, carry=0):
    dfpload6502(DFPREG_A, num)
    regs, _ = jsr(dfp_addresses[9], x=DFPREG_A, p=1 if carry else 0)
    return dfpstore6502(DFPREG_A), regs['p'] & 0x01

def dfpinc6502(num):
    dfpload6502(DFPREG_A, num)
    regs, _ = jsr(dfp_addresses[10], x=DFPREG_A)
    return dfpstore6502(DFPREG_A), regs['p'] & 0x01

def dfpadd6502(a, b):
    dfpload6502(DFPREG_A, a)
    dfpload6502(DFPREG_B, b)
    jsr(dfp_addresses[12], x=DFPREG_B)
    return dfpstore6502(DFPREG_A)

def dfpsub6502(a, b, signonly=0):
    dfpload6502(DFPREG_A, a)
    dfpload6502(DFPREG_B, b)
    regs, _ = jsr(dfp_addresses[11 if signonly else 13], x=DFPREG_B)
    sign = regs['a']
    if sign >= 128: sign -= 256
    if signonly:
        return sign
    return dfpstore6502(DFPREG_A), sign

def dfpmul6502(a, b, signonly=0):
    dfpload6502(DFPREG_A, a)
    dfpload6502(DFPREG_B, b)
    regs, c = jsr(dfp_addresses[14], x=DFPREG_B,
                  trace=False)
    # usually c is 4600
    return dfpstore6502(DFPREG_A)

def test_dfpfmt():
    print(floatypoint.dfpmake(11))
    i = floatypoint.dfpmake(12345)
    i = floatypoint.dfpmake(1234500, -1)
    print(floatypoint.dfpfmt(i))
    print(dfpfmt6502(i))
    print("nzrange of %s is %s on pydfp or %s on dfp6502"
          % (i, floatypoint.dfpnzrange(i), dfpnzrange6502(i)))
    for num in ((11,), (4440000,), (12345678987654321, 0),
                (11, -1), (11, -2)):
        num = floatypoint.dfpmake(*num)
        print("%s (py) = %s (6502)"
              % (floatypoint.dfpfmt(num), dfpfmt6502(num)))
    for i in (123, 1234, 12345, 123456, 1234567, 12345678, 123456789, 1234567890, 12345678901):
        i = floatypoint.dfpmake(i)
        rows = [(length,
                 floatypoint.dfpfmt_metric(i, length),
                 dfpfmt_metric6502(i, length))
                for length in xrange(6, 15)]
        if any(b != c for (a, b, c) in rows):
            print("beautify failure!")
        print("\n".join("%2d: %14s (py) = %14s (6502)" % row
                        for row in rows))

def test_dfpadd():
    amt_to_add = floatypoint.dfpmake(987650, -2)
    for i in (123, 1234, 12345, 123456, 1234567, 12345678, 123456789, 1234567890, 12345678901):
        native_sum = i + 98.765
        i = floatypoint.dfpmake(i)
        py_sum = floatypoint.dfpadd(i, amt_to_add)
        nes_sum = floatypoint.dfpadd(i, amt_to_add)
        if nes_sum != py_sum:
            print("add failure! %s (py) != %s (6502)"
                  % (floatypoint.dfpfmt(py_sum), floatypoint.dfpfmt(nes_sum)))
        print("%15.3f %15s" % (native_sum, floatypoint.dfpfmt(py_sum)))

def test_dfpinc():
    nums = [
        (23,99,99,99,99,15,0x80),
        (23,99,99,99,99,99,0x81),
        (23,99,99,64,99,99,0x81),
        (57,56,57,56,57,56,0x81),
        ( 0, 0, 0, 0, 0, 1,0x81),
        (99,99,99,99,99,99,0x81),
    ]
    funcs = [
        ("add one ulp", floatypoint.dfpinc, dfpinc6502),
        ("multiply by 2", floatypoint.dfpdouble, dfpdouble6502),
        ("divide by 2", floatypoint.dfphalve, dfphalve6502),
    ]
    for name, pyfunc, nesfunc in funcs:
        for n in nums:
            nresult, ncarry = nesfunc(n)
            ls = list(n)
            # pyfunc operates in place.  Well so does nesfunc, but
            # the 6502 test harness abstracts that.
            pycarry = pyfunc(ls)
            pyresult = tuple(ls)
            if pyresult != nresult or pycarry != ncarry:
                print("FAIL!")
            print("%s %s\n  python: %s c%d\n    6502: %s c%d"
                  % (repr(n), name,
                     repr(pyresult), pycarry,
                     repr(nresult), ncarry))

def test_dfpsub():
    subtrahend = floatypoint.dfpmake(45678, -1)
    for i in (123, 1234, 12345, 123456, 1234567, 12345678, 123456789, 1234567890, 12345678901):
        print(i - 456.78)
        minuend = floatypoint.dfpmake(i)
        pydiff, pysign = floatypoint.dfpsub(minuend, subtrahend)
        nesdiff, nessign = dfpsub6502(minuend, subtrahend)
        print("nessign is %02x" % nessign)
        if pydiff != nesdiff or pysign != nessign:
            print("Fail?")
            print("%s != %s" % (repr(pydiff), repr(nesdiff)))
        tpysign = "0+-"[pysign]
        tnessign = "0+-"[nessign]
        print("%s%s %s%s"
              % (tpysign, floatypoint.dfpfmt_metric(pydiff, 15),
                 tnessign, floatypoint.dfpfmt_metric(nesdiff, 15)))

    # Rounding and guard digit behavior to be implemented and
    # tested later
    minuend = floatypoint.dfpmake(1234567)
    subtrahend = floatypoint.dfpmake(234567)
    print(dfpsub6502(minuend, subtrahend))
    subtrahend = floatypoint.dfpmake(234568)
    print(dfpsub6502(minuend, subtrahend))

def dfpdump(n):
    c = ["%02d" % dp for dp in n[:-1]]
    c.append("$%02x" % n[-1])
    return ''.join(c)

def test_dfpmul():
    from itertools import chain
    pairs = [
        (2, 3), (22, 33), (101, 1), (12345, 54321),
        (174900625, 1), (1, 174900625), (174900625, 174900625),
        ((2, 4), (3, -4)), ((2, -60), (3, -60)), ((2, -70), (3, -70)),
    ]
    fmt = floatypoint.dfpfmt
    for a, b in pairs:
        if not isinstance(a, tuple):
            a = (a, 0)
        two = floatypoint.dfpmake(*a)
        a = a[0] * (100 ** a[1])
        if not isinstance(b, tuple):
            b = (b, 0)
        three = floatypoint.dfpmake(*b)
        b = b[0] * (100.0 ** b[1])
        six = floatypoint.dfpmul(two, three)
        nsix = dfpmul6502(two, three)
        ttwo, tthree, tsix, tnsix = (fmt(i) for i in (two, three, six, nsix))
        print("%g * %g = %g\n   py: %s * %s = %s\n  nes: %s * %s = %s"
              % (a, b, a * b, ttwo, tthree, tsix, ttwo, tthree, tnsix))

    print("\nCost of cursors in Cookie Clicker")
    bcost = floatypoint.calc_building_cost
    base_cost = floatypoint.dfpmake(15)
    nums = chain(xrange(0, 240, 5), xrange(240, 247))
    for num in nums:
        cost = bcost(base_cost, num)
        ncost = bcost(base_cost, num, mul=dfpmul6502)
        print("Cursor while owning %d should cost %d cookies\n(py: %s; NES: %s)"
              % (num, 15 * 1.15 ** num,
                 fmt(cost), fmt(ncost)))
    print("246th cursor in game costs 12,561,178,154,785,838 with\n"
          "both Santa upgrades that affect building prices")
    print("(py: %s; NES: %s)"
          % (fmt(bcost(base_cost, num, 2)),
             fmt(bcost(base_cost, num, 2, mul=dfpmul6502))))


def main():
    global dfp_addresses
    args = ['make', 'floattest.prg', '-j2']
    subprocess.check_call(args, cwd='..')
    with open("../floattest.prg", "rb") as infp:
        sim.rom.fromstring(infp.read())
    dfp_addresses = [sim.rom[i] + 256 * sim.rom[i + 1]
                     for i in xrange(-64, 0, 2)]

##    test_dfpfmt()
##    test_dfpadd()
##    test_dfpinc()
##    test_dfpsub()
    test_dfpmul()



if __name__=='__main__':
    main()
