#!/usr/bin/env python
from __future__ import with_statement, print_function, division
from vwfbuild import vwfcvt
import sys, itertools
from collections import defaultdict

# The news ticker occupies the bottom 2 lines.
# Each ticker entry is broken into two lines, each no longer than
# 192 pixels. 
# '" --', '; ', ', ', ' - ', ' "'


# Some strings are preferred break points.
# word before space ends with this string,
# word after space starts with this string,
# True to prevent single-line returns if possible
good_breaks = [
    ('"', '--', True),   # grandma quotes
    (';', '', False),    # clause splice
    (',', '', False),    # conjunction splice
    ('.', '', False),    # sentence splice
    ('', '"', False)
]

def strwidth(word, char_widths):
    return sum(char_widths[ord(c) - 32] for c in word)

def linebreak_pied(wordwidths, linewidth):
    """Fit a list of words to lines no wider than linewidth.

The function uses two techniques to ensure a tight fit.
Pied-piping moves the last word on a line down to the next line if
the part of the line before the word is shorter than the next line.
Narrowing reperforms the entire layout process using a width one
pixel less than before, which breaks up a local optimum with a long
line early in a paragraph.

wordwidths and linewidth are supposed to include one space width.

"""

    best = None
    while True:
        pxcounts = [0]
        breaks = []  # index of the first word on the following line

        # Lay out as many words as possible on early lines
        for i, w in enumerate(wordwidths):
            pxcount = pxcounts[-1] + w
            if pxcount > linewidth:
                pxcounts.append(w)
                breaks.append(i)
            else:
                pxcounts[-1] = pxcount

        # Try moving last word on each line to the next line
        i = len(breaks) - 1
        while i >= 0:
            # The last word on the line is wordwidths[breaks[i] - 1].
            # If it can be moved to the next line, do so.
            w = wordwidths[breaks[i] - 1]
            if pxcounts[i] > pxcounts[i + 1] + w:
                pxcounts[i] -= w
                pxcounts[i + 1] += w
                breaks[i] -= 1
                # Pied-pipe it to the next line if possible
                i = min(i + 1, len(breaks) - 1)
            else:
                # Last word cannot be moved, so try an earlier line
                i -= 1

        # Try narrowing the layout further.  If narrowing increases
        # line count, return the previous best.  Narrowing won't
        # improve the fit very often except in fairly long texts
        # with a long first line.
        if best is None or len(pxcounts) <= len(best[0]):
            if best is not None:
                print("Successful narrowing! %s to %s"
                      % (repr(best[0]), repr(pxcounts)))
            best = (pxcounts, breaks)
            if len(pxcounts) < 2:
                return best
            linewidth = max(pxcounts) - 1
        else:
            return best
        # TO DO: Optimize narrowing by moving only words on lines
        # that exceed the new maximum line width to the next line.


def breaks_to_slices(breaks, length):
    slices = [(breaks[i - 1] if i else 0, breaks[i])
              for i in xrange(len(breaks))]
    slices.append((slices[-1][1] if slices else 0, length))
    return slices

def ordered_dedupe(items):
    """Iterate over the first occurrence of each hashable value in an iterable."""
    seen = set()
    for el in items:
        if el not in seen:
            yield el
            seen.add(el)

def linebreak_with_breaks(breaks, wordwidths, linewidth, numlines, words):
    segs = breaks_to_slices(sorted(breaks), len(wordwidths))
    candidate_break = [linebreak_pied(wordwidths[s:e], linewidth)
                       for s, e in segs]
    total_lines = sum(len(row[0]) for row in candidate_break)
    if total_lines <= numlines:
        c = zip(segs, (row[1] for row in candidate_break))
        c = [' '.join(words[s + s2:s + e2])
             for ((s, e), b) in c
             for (s2, e2) in breaks_to_slices(b, e - s)]
        return c

def linebreak(words, char_widths, numlines, linewidth):
    spacewidth = char_widths[0]
    linewidth += spacewidth
    wordwidths = [strwidth(word, char_widths) + spacewidth for word in words]
    maxwordwidth = max(wordwidths)
    if maxwordwidth > linewidth:
        raise ValueError("word too long: %d pixels"
                         % (maxwordwidth - spacewidth))
    msgwidth = sum(wordwidths)
    if msgwidth > numlines * linewidth:
        raise ValueError("message too long: %d pixels"
                         % (msgwidth - spacewidth))

    # Try to find preferred line break points
    found_good_breaks = []
    found_firm_breaks = set()
    for brkid, (ends1, starts2, firm) in enumerate(good_breaks):
        for i in xrange(1, len(wordwidths)):
            if words[i - 1].endswith(ends1) and words[i].startswith(starts2):
                found_good_breaks.append(i)
                if firm: found_firm_breaks.add(i)

    # If the message is short enough for one line and no found break
    # points are attractive enough to rule out a single line, succeed
    if msgwidth <= linewidth and not found_firm_breaks:
        return [' '.join(words)]

    # Lower bound on how many lines are required.  This is the last
    # thing that can cause a message to fail to fit
    base_pxcounts, base_breaks = linebreak_pied(wordwidths, linewidth)
    if len(base_pxcounts) > numlines:
        raise ValueError("message too long: %d lines" % len(base_pxcounts))

    # If firm breaks exist, the line count is allowed to exceed
    # the lower bound but not the cap.  Otherwise, cap line count
    # at the lower bound.
    if not found_firm_breaks:
        numlines = len(base_pxcounts)

    # If firm breaks exist, try breaking at every firm break,
    # and see if it comes under the line cap.  If so, break
    # at all firm breaks and ignore other preferred breaks.
    
    if 0 < len(found_firm_breaks) <= numlines:
        c = linebreak_with_breaks(found_firm_breaks, wordwidths, linewidth, numlines, words)
        if c is not None:
            return c

    # Try breaking at each preferred break to see if that comes in
    # under the line cap.
    combis = [frozenset(c)
              for length in xrange(len(found_good_breaks), -1, -1)
              for c in itertools.combinations(found_good_breaks, length)]
    for breaks in ordered_dedupe(combis):
        c = linebreak_with_breaks(breaks, wordwidths, linewidth, numlines, words)
        if c is not None:
            return c
    assert False

def main():
    wordhisto = defaultdict(lambda: 0)

    char_widths = vwfcvt('../tilesets/vwf7.png')[0]
    with open('../docs/ticker.txt') as infp:
        body = infp.readlines()
    foundletsmake = False
    num_failures = 0
    errlines = []
    for line in body:
        words = line.split()
        if not words:
            continue
        if words[0].startswith('['):
            continue
        if not foundletsmake:
            if len(words) >= 2 and words[0] == "Let's" and words[1] == "make":
                foundletsmake = True
            continue
        eqsplit = words[0].split('=', 1)
        if len(eqsplit) > 1 and eqsplit[0].isdigit():
            if eqsplit[1]:
                words[0] = eqsplit[1]
            else:
                del words[0]
        try:
            lines = linebreak(words, char_widths, 2, 192)
        except ValueError as e:
            errlines.append(' '.join(words))
            errlines.append(str(e))
            num_failures += 1
        else:
            for word in words:
                wordhisto[word] += 1
            if len(errlines) > 0:
                print("\n".join(errlines), file=sys.stderr)
                del errlines[:]
            if lines:
                pass
    sortedhisto = sorted(wordhisto.iteritems(), key=lambda x: -x[1])
    print("%d distinct words" % len(sortedhisto))
    print("\n".join(repr(row) for row in sortedhisto[:20]))

    if len(errlines) > 0:
        print("\n".join(errlines), file=sys.stderr)
    if num_failures > 0:
        print("%d lines not processed" % num_failures)
##                print("%d:\n  %s" % (len(lines), "\n".join(lines)))

if __name__=='__main__':
    main()
